@@ -11,11 +11,14 @@ from rich.tree import Tree
 from rich.panel import Panel
 from rich.layout import Layout
 from rich.columns import Columns
+import requests
+from enum import Enum
 #Local Imports
 from ..dca import Arps, Wor, FreqEnum, Forecast, converter_factor
 from ..cashflow import CashFlowModel, CashFlow, CashFlowParams, ChgPts, npv_cashflows, irr_cashflows
 from ..console import console
-
+from ..auth import Credential
+import traceback
 # Put together all classes of DCA in a Union type. Pydantic uses this type to validate
 # the input dca is a subclass of DCA. 
 # Still I don't know if there's a way Pydantic check if a input variable is subclass of other class
@@ -28,13 +31,20 @@ freq_format={
     'A':'%Y'
 }
 
+class SchemasEnum(str, Enum):
+    period = 'period'
+    scenario = 'scenario'
+    well = 'well'
+    wellsgroup = 'wellsgroup'
+
+
 class Depends(BaseModel):
     period : str = Field(...)
     delay : Union[timedelta,int] = Field(None)
   
     
 class ScheduleBase(BaseModel):
-	name:str
+	name:str = Field(None)
 	cashflow_params : Optional[List[CashFlowParams]] = Field(None)
 	cashflow : Optional[List[CashFlowModel]] = Field(None)
 	forecast: Optional[Forecast] = Field(None)
@@ -42,6 +52,7 @@ class ScheduleBase(BaseModel):
 	iter : int = Field(1, ge=1)
 	ppf : Optional[float] = Field(None, ge=0, le=1)
 	description: str = Field(None)
+	id: str = Field(None)
  
 	class Config:
 		arbitrary_types_allowed = True
@@ -61,29 +72,112 @@ class ScheduleBase(BaseModel):
 	def to_file(self, file:str, format='yaml'):
 		with open(f'{file}.{format}','w') as f:
 			if format=='yaml':
-				yaml.safe_dump(json.loads(self.json(exclude_unset=True)), f)
+				yaml.safe_dump(json.loads(self.json(exclude_none=True)), f)
 			if format=='json':
-				f.write(self.json(exclude_unset=True))
+				f.write(self.json(exclude_none=True))
     
 	#def tree(self):
 	#	node_tree = Tree(self.name)
+	def get_db(self,key:str, cred:Credential):
+		end_point = f'api/v1/models/{self.type.value}/'
+		headers = {
+			'accept': 'application/json',
+			'Authorization': f'Bearer {cred.token}'
+		}	
+		try:
+			r = requests.get(f'{cred.url}{end_point}{key}', headers=headers)
+			r.raise_for_status()
+			data = json.loads(r.text)
+			dict_data = data['case']
+			for i in dict_data:
+				setattr(self,i,dict_data[i]) 
+		except requests.exceptions.HTTPError as err:
+			print(err)
+		else:
+			self.id = key 
+  
+	def insert_db(self,cred:Credential, description:str=None):
+		end_point = f'api/v1/models/{self.type.value}'
+		headers = {
+			'accept': 'application/json',
+			'Authorization': f'Bearer {cred.token}',
+			'Content-Type':'application/json'
+		}
+		model = json.loads(self.json(exclude_unset=True))
+		data = {'model':model}
+		if description:
+			data['description'] = description
+		try:
+			r = requests.post(f'{cred.url}{end_point}', headers=headers, json=data)
+			r.raise_for_status()
+			data = json.loads(r.text)
+			
+		except requests.exceptions.HTTPError as err:
+			print(err)
+		else:
+			self.id = data['id']
+			return data['id']
+
+	def update_db(self, cred:Credential, description:str=None):
+		if self.id is None:
+			raise ValueError('Model has no Key')
+		end_point = f'api/v1/models/{self.type.value}/'
+		headers = {
+			'accept': 'application/json',
+			'Authorization': f'Bearer {cred.token}'
+		}
+		model = json.loads(self.json(exclude_unset=True))
+		data = {'model':model}
+		if description:
+			data['description'] = description
+		try:
+			r = requests.put(f'{cred.url}{end_point}{self.id}', headers=headers, json=data)
+			r.raise_for_status()
+			data = json.loads(r.text)
+		except requests.exceptions.HTTPError as err:
+			print(err)
+		else:
+			return data['id']
+			
+   
+	def delete_db(self, cred:Credential, description:str=None):
+		if self.id is None:
+			raise ValueError('Model has no Key')
+		end_point = f'api/v1/models{self.type.value}/'
+		headers = {
+			'accept': 'application/json',
+			'Authorization': f'Bearer {cred.token}'
+		}
+		model = json.loads(self.json(exclude_unset=True))
+		data = {'model':model}
+		if description:
+			data['description'] = description
+		try:
+			r = requests.delete(f'{cred.url}{end_point}{self.id}', headers=headers, json=data)
+			r.raise_for_status()
+			data = json.loads(r.text)
+		except requests.exceptions.HTTPError as err:
+			print(err)
+		else:
+			return data['id']
   
 class Period(ScheduleBase):
-	dca : union_classes_dca 
-	start: Union[int,date]
-	end: Optional[Union[int,date]]
+	dca : union_classes_dca = Field(None)
+	start: Union[int,date] = Field(None)
+	end: Optional[Union[int,date]] = Field(None)
 	time_list : Optional[List[Union[int,date]]] = Field(None)
 	freq_input: FreqEnum = Field('D')
 	freq_output: FreqEnum = Field('D')
 	rate_limit: Optional[float] = Field(None, ge=0)
 	cum_limit: Optional[float] = Field(None, ge=0)
 	depends: Optional[Depends] = Field(None)
+	type: SchemasEnum = Field(SchemasEnum.period, const=True)
 
-	@validator('end')
-	def start_end_match_type(cls,v,values):
-	    if type(v) != type(values['start']):
-	        raise ValueError('start and end must be the same type')
-	    return v
+	# @validator('end')
+	# def start_end_match_type(cls,v,values):
+	#     if type(v) != type(values['start']):
+	#         raise ValueError('start and end must be the same type')
+	#     return v
 
 	class Config:
 		arbitrary_types_allowed = True
@@ -146,8 +240,8 @@ class Period(ScheduleBase):
 
 		if ppf is None:
 			ppf = self.ppf
-   
-		if self.forecast is not None and self.cashflow_params is not None:
+
+		if self.forecast is not None and any([self.cashflow_params is not None,add_cash_params is not None]):
 
 			_forecast = self.forecast.df()
 
@@ -165,21 +259,13 @@ class Period(ScheduleBase):
 			forecast_iterations = _forecast['iteration'].unique()
 			shapes_to_broadcast.append(len(forecast_iterations))
 
-			#Cashflows Iterations
-			for p in self.cashflow_params:
-				shapes_to_broadcast.append(p.iter)
-    
-			#shapes broadcast
-			shapes = np.broadcast_shapes(*shapes_to_broadcast)
-		
 			#Forecast iterations shape.
 			#Example. If there's only one iteration in Forecast 
 			# the variable forecast_iterations would be = np.array([0]). 
 			# At the same time if the broadcasted shape is (10,) due to 10 iterations in cashflows params
 			# the iterate_new_shape would be np.array([0,0,0,0,0,0,0,0,0,0]). 
 			# This is done to make the 10 different cashflow models.
-			iterate_new_shape = forecast_iterations * np.ones(shapes)
-   
+			
 			#Iterate over list of cases
 			if self.cashflow_params is None:
 				cashflow_params = []
@@ -187,6 +273,14 @@ class Period(ScheduleBase):
 				cashflow_params = self.cashflow_params.copy()
 			if add_cash_params:
 				cashflow_params.extend(add_cash_params)
+
+			#Cashflows Iterations
+			for p in cashflow_params:
+				shapes_to_broadcast.append(p.iter)
+    
+			#shapes broadcast
+			shapes = np.broadcast_shapes(*shapes_to_broadcast)
+			iterate_new_shape = forecast_iterations * np.ones(shapes)
     
 			if len(cashflow_params)==0:
 				raise ValueError('No Cashflow Params are set')
@@ -298,17 +392,18 @@ class Period(ScheduleBase):
 		return node_tree
 
 	def layout(self, emoji=':chart_with_downwards_trend:', title_style = 'bold green'):
-		text = yaml.dump(self.dict(exclude_unset=True))  
+		text = yaml.dump(self.dict(exclude_unset=True, exclude_none=True))  
 
 		panel_text = f'{emoji}\n' + text
 		panel = Panel(panel_text,title=f'[{title_style}]{self.name}[/{title_style}]')
 		return panel
-  
+
 class Scenario(ScheduleBase):
-	periods: Union[List[Period],Dict[str,Period]]
+	periods: Union[List[Period],Dict[str,Period]] = Field(None)
 	freq_output: str = Field('D')
+	type: SchemasEnum = Field(SchemasEnum.scenario, const=True)
  
-	@validator('periods', always=True)
+	@validator('periods')
 	def match_periods_freqs(cls,v):
 		if isinstance(v,list):
 			v = {i.name:i for i in v}
@@ -445,7 +540,8 @@ class Scenario(ScheduleBase):
 					csh_name = add_name + '-' + self.name
 				_cf = self.periods[p].generate_cashflow(freq_output=freq_output, add_name=csh_name, seed=seed, ppf=ppf, add_cash_params=pass_cashflow_params)
 			except Exception as e:
-				print(e)
+				print(p,e)
+				traceback.print_exc()
 				list_periods_errors.append(self.periods[p].name)
 			else:
 				if len(_cf)==1:
@@ -540,13 +636,13 @@ class Scenario(ScheduleBase):
 			return panel
 
 class Well(ScheduleBase):
-	scenarios : Union[List[Scenario],Dict[str,Scenario]]
- 
+	scenarios : Union[List[Scenario],Dict[str,Scenario]] = Field(None)
+	type: SchemasEnum = Field(SchemasEnum.well, const=True)
 	class Config:
 		arbitrary_types_allowed = True
 		validate_assignment = True
 
-	@validator('scenarios', always=True)
+	@validator('scenarios')
 	def match_periods_freqs(cls,v):
 		if isinstance(v,list):
 			v = {i.name:i for i in v}
@@ -723,13 +819,14 @@ class Well(ScheduleBase):
 
    
 class WellsGroup(ScheduleBase):
-	wells : Union[List[Well],Dict[str,Well]]
+	wells : Union[List[Well],Dict[str,Well]] = Field(None)
+	type: SchemasEnum = Field(SchemasEnum.wellsgroup, const=True)
 
 	class Config:
 		arbitrary_types_allowed = True
 		validate_assignment = True
 
-	@validator('wells', always=True)
+	@validator('wells')
 	def match_periods_freqs(cls,v):
 		if isinstance(v,list):
 			v = {i.name:i for i in v}
